"""
Utilities that don't go anywhere else.
"""

from __future__ import unicode_literals

import sys
from types import ModuleType

from six import exec_, text_type as unicode, PY3


def safeunicode(o):
    """
    Like C{unicode()}, but catches and swallows any raised exceptions.

    @param o: An object of some sort.

    @return: C{unicode(o)}, or an error message if that failed.
    @rtype: C{unicode}
    """
    try:
        return unicode(o)
    except:
        # Not much we can do about this...
        return "eliot: unknown, unicode() raised exception"


def saferepr(o):
    """
    Like C{unicode(repr())}, but catches and swallows any raised exceptions.

    @param o: An object of some sort.

    @return: C{unicode(repr(o))}, or an error message if that failed.
    @rtype: C{unicode}
    """
    try:
        return unicode(repr(o))
    except:
        # Not much we can do about this...
        return "eliot: unknown, unicode() raised exception"


def load_module(name, original_module):
    """
    Load a copy of a module, distinct from what you'd get if you imported
    it directly.

    @param str name: The name of the new module.
    @param original_module: The original module we're recreating.

    @return: A new, distinct module.
    """
    module = ModuleType(name)
    if PY3:
        import importlib.util

        spec = importlib.util.find_spec(original_module.__name__)
        source = spec.loader.get_code(original_module.__name__)
    else:
        if getattr(sys, "frozen", False):
            raise NotImplementedError("Can't load modules on Python 2 with PyInstaller")
        path = original_module.__file__
        if path.endswith(".pyc") or path.endswith(".pyo"):
            path = path[:-1]
        with open(path) as f:
            source = f.read()
    exec_(source, module.__dict__, module.__dict__)
    return module



"""
Standardized serialization code.
"""

from __future__ import unicode_literals

from hashlib import md5

_TIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"


def timestamp(dt):
    """
    Convert a UTC datetime to a string.

    @param dt: A C{datetime.datetime} in UTC timezone.

    @return: C{unicode}
    """
    return dt.strftime(_TIME_FORMAT)


def identity(value):
    """
    Return the passed in object.
    """
    return value


def md5hex(data):
    """
    Return hex MD5 of the input bytes.

    @param data: Some C{bytes}.

    @return: Hex-encoded MD5 of the data.
    """
    return md5(data).hexdigest()



"""
Tests for coroutines.

Imported into test_coroutine.py when running tests under Python 3.5 or later;
in earlier versions of Python this code is a syntax error.
"""

import asyncio
from unittest import TestCase

from ..testing import capture_logging
from ..parse import Parser
from .. import start_action


async def standalone_coro():
    """
    Log a message inside a new coroutine.
    """
    await asyncio.sleep(0.1)
    with start_action(action_type="standalone"):
        pass


async def calling_coro():
    """
    Log an action inside a coroutine, and call another coroutine.
    """
    with start_action(action_type="calling"):
        await standalone_coro()


def run_coroutines(*async_functions):
    """
    Run a coroutine until it finishes.
    """
    loop = asyncio.get_event_loop()
    futures = [asyncio.ensure_future(f()) for f in async_functions]

    async def wait_for_futures():
        for future in futures:
            await future

    loop.run_until_complete(wait_for_futures())


class CoroutineTests(TestCase):
    """
    Tests for coroutines.
    """

    @capture_logging(None)
    def test_multiple_coroutines_contexts(self, logger):
        """
        Each top-level coroutine has its own Eliot logging context.
        """

        async def waiting_coro():
            with start_action(action_type="waiting"):
                await asyncio.sleep(0.5)

        run_coroutines(waiting_coro, standalone_coro)
        trees = Parser.parse_stream(logger.messages)
        self.assertEqual(
            sorted([(t.root().action_type, t.root().children) for t in trees]),
            [("standalone", []), ("waiting", [])],
        )

    @capture_logging(None)
    def test_await_inherits_coroutine_contexts(self, logger):
        """
        awaited coroutines inherit the logging context.
        """
        run_coroutines(calling_coro)
        [tree] = Parser.parse_stream(logger.messages)
        root = tree.root()
        [child] = root.children
        self.assertEqual(
            (root.action_type, child.action_type, child.children),
            ("calling", "standalone", []),
        )

    @capture_logging(None)
    def test_interleaved_coroutines(self, logger):
        """
        start_action() started in one coroutine doesn't impact another in a
        different coroutine.
        """

        async def coro_sleep(delay, action_type):
            with start_action(action_type=action_type):
                await asyncio.sleep(delay)

        async def main():
            with start_action(action_type="main"):
                f1 = asyncio.ensure_future(coro_sleep(1, "a"))
                f2 = asyncio.ensure_future(coro_sleep(0.5, "b"))
                await f1
                await f2

        run_coroutines(main)
        [tree] = list(Parser.parse_stream(logger.messages))
        root = tree.root()
        self.assertEqual(root.action_type, "main")
        self.assertEqual(sorted([c.action_type for c in root.children]), ["a", "b"])



"""
Tests for L{eliot.serializers}.
"""

from __future__ import unicode_literals

from unittest import TestCase
from datetime import datetime
from hashlib import md5

from ..serializers import timestamp, identity, md5hex


class SerializerTests(TestCase):
    """
    Tests for standard serializers.
    """

    def test_timestamp(self):
        """
        L{timestamp} converts a UTC L{datetime} to a Unicode strings.
        """
        dt = datetime(2012, 9, 28, 14, 53, 6, 123456)
        self.assertEqual(timestamp(dt), "2012-09-28T14:53:06.123456Z")

    def test_identity(self):
        """
        L{identity} returns the input object.
        """
        obj = object()
        self.assertIs(identity(obj), obj)

    def test_md5hex(self):
        """
        L{md5hex} returns the hex value of a MD5 checksum.
        """
        data = b"01234456789"
        self.assertEqual(md5hex(data), md5(data).hexdigest())



"""
Logging of tracebacks and L{twisted.python.failure.Failure} instances,
as well as common utilities for handling exception logging.
"""

from __future__ import unicode_literals

import traceback
import sys

from ._message import EXCEPTION_FIELD, REASON_FIELD
from ._util import safeunicode, load_module
from ._validation import MessageType, Field
from ._errors import _error_extraction

TRACEBACK_MESSAGE = MessageType(
    "eliot:traceback",
    [
        Field(REASON_FIELD, safeunicode, "The exception's value."),
        Field("traceback", safeunicode, "The traceback."),
        Field(
            EXCEPTION_FIELD,
            lambda typ: "%s.%s" % (typ.__module__, typ.__name__),
            "The exception type's FQPN.",
        ),
    ],
    "An unexpected exception indicating a bug.",
)
# The fields here are actually subset of what you might get in practice,
# due to exception extraction, so we hackily modify the serializer:
TRACEBACK_MESSAGE._serializer.allow_additional_fields = True


def _writeTracebackMessage(logger, typ, exception, traceback):
    """
    Write a traceback to the log.

    @param typ: The class of the exception.

    @param exception: The L{Exception} instance.

    @param traceback: The traceback, a C{str}.
    """
    msg = TRACEBACK_MESSAGE(reason=exception, traceback=traceback, exception=typ)
    msg = msg.bind(**_error_extraction.get_fields_for_exception(logger, exception))
    msg.write(logger)


# The default Python standard library traceback.py formatting functions
# involving reading source from disk. This is a potential performance hit
# since disk I/O can block. We therefore format the tracebacks with in-memory
# information only.
#
# Unfortunately, the easiest way to do this is... exciting.
def _get_traceback_no_io():
    """
    Return a version of L{traceback} that doesn't do I/O.
    """
    try:
        module = load_module(str("_traceback_no_io"), traceback)
    except NotImplementedError:
        # Can't fix the I/O problem, oh well:
        return traceback

    class FakeLineCache(object):
        def checkcache(self, *args, **kwargs):
            None

        def getline(self, *args, **kwargs):
            return ""

        def lazycache(self, *args, **kwargs):
            return None

    module.linecache = FakeLineCache()
    return module


_traceback_no_io = _get_traceback_no_io()


def write_traceback(logger=None, exc_info=None):
    """
    Write the latest traceback to the log.

    This should be used inside an C{except} block. For example:

         try:
             dostuff()
         except:
             write_traceback(logger)

    Or you can pass the result of C{sys.exc_info()} to the C{exc_info}
    parameter.
    """
    if exc_info is None:
        exc_info = sys.exc_info()
    typ, exception, tb = exc_info
    traceback = "".join(_traceback_no_io.format_exception(typ, exception, tb))
    _writeTracebackMessage(logger, typ, exception, traceback)


def writeFailure(failure, logger=None):
    """
    Write a L{twisted.python.failure.Failure} to the log.

    This is for situations where you got an unexpected exception and want to
    log a traceback. For example, if you have C{Deferred} that might error,
    you'll want to wrap it with a L{eliot.twisted.DeferredContext} and then add
    C{writeFailure} as the error handler to get the traceback logged:

        d = DeferredContext(dostuff())
        d.addCallback(process)
        # Final error handler.
        d.addErrback(writeFailure)

    @param failure: L{Failure} to write to the log.

    @type logger: L{eliot.ILogger}. Will be deprecated at some point, so just
        ignore it.

    @return: None
    """
    # Failure.getBriefTraceback does not include source code, so does not do
    # I/O.
    _writeTracebackMessage(
        logger, failure.value.__class__, failure.value, failure.getBriefTraceback()
    )



"""
A log destination for use by Twisted applications.

Runs in a thread, so that we don't do blocking I/O in the event loop thread.
"""

from __future__ import unicode_literals, absolute_import

import threading
import select
from warnings import warn

from twisted.application.service import Service
from twisted.internet.threads import deferToThreadPool

if getattr(select, "poll", None):
    from twisted.internet.pollreactor import PollReactor as Reactor
else:
    from twisted.internet.selectreactor import SelectReactor as Reactor

from . import addDestination, removeDestination
from ._output import FileDestination


class ThreadedWriter(Service):
    """
    An non-blocking Eliot log destination that wraps a blocking
    destination, writing log messages to the latter in a managed thread.

    Unfortunately Python's Queue is not reentrant
    (http://bugs.python.org/issue14976) and neither is RLock
    (http://bugs.python.org/issue13697). In order to queue items in a thread we
    therefore rely on the self-pipe trick, and the easiest way to do that is by
    running another reactor in the thread.

    @ivar _reactor: A private reactor running in a thread which will do the log
        writes.

    @ivar _thread: C{None}, or a L{threading.Thread} running the private
        reactor.
    """

    name = "Eliot Log Writer"

    def __init__(self, destination, reactor):
        """
        @param destination: The underlying destination for log files. This will
            be called from a non-reactor thread.

        @param reactor: The main reactor.
        """
        self._destination = destination
        self._reactor = Reactor()
        # Ick. See https://twistedmatrix.com/trac/ticket/6982 for real solution.
        self._reactor._registerAsIOThread = False
        self._mainReactor = reactor
        self._thread = None

    def startService(self):
        """
        Start the writer thread.
        """
        Service.startService(self)
        self._thread = threading.Thread(target=self._writer)
        self._thread.start()
        addDestination(self)

    def stopService(self):
        """
        Stop the writer thread, wait for it to finish.
        """
        Service.stopService(self)
        removeDestination(self)
        self._reactor.callFromThread(self._reactor.stop)
        return deferToThreadPool(
            self._mainReactor, self._mainReactor.getThreadPool(), self._thread.join
        )

    def __call__(self, data):
        """
        Add the data to the queue, to be serialized to JSON and written by the
        writer thread with a newline added.

        @param data: C{bytes} to write to disk.
        """
        self._reactor.callFromThread(self._destination, data)

    def _writer(self):
        """
        The function run by the writer thread.
        """
        self._reactor.run(installSignalHandlers=False)


class ThreadedFileWriter(ThreadedWriter):
    """
    ``ThreadedWriter`` that takes a log file and writes to it using a
    ``FileDestination``.

    This exists for backwards compatibility purpose. The recommended API is
    ``ThreadedWriter``.
    """

    def __init__(self, logFile, reactor):
        """
        @param logFile: A C{file}-like object that is at the end of its
            existing contents (e.g. opened with append mode) and accepts
            bytes.
        @type logFile: C{file}, or any file-like object with C{write},
            C{flush} and C{close} methods e.g. a
            L{twisted.python.logfile.LogFile} if you want log rotation.

        @param reactor: The main reactor.
        """
        warn(
            "ThreadedFileWriter is deprecated since 0.9.0. "
            "Use ThreadedWriter instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self._logFile = logFile
        ThreadedWriter.__init__(self, FileDestination(file=logFile), reactor)

    def stopService(self):
        d = ThreadedWriter.stopService(self)
        d.addCallback(lambda _: self._logFile.close())
        return d



"""
Eliot: Logging for Complex & Distributed Systems.
"""
from warnings import warn
from sys import version_info

# Enable asyncio contextvars support in Python 3.5/3.6:
if version_info < (3, 7):
    # On Python 3.5.2 and earlier, some of the necessary attributes aren't exposed:
    if version_info < (3, 5, 3):
        raise RuntimeError(
            "This version of Eliot doesn't work on Python 3.5.2 or earlier. "
            "Either upgrade to Python 3.5.3 or later (on Ubuntu 16.04 "
            "you can use https://launchpad.net/~deadsnakes/+archive/ubuntu/ppa "
            "to get Python 3.6), or pin Eliot to version 1.7."
        )
    import aiocontextvars

    dir(aiocontextvars)  # pacify pyflakes
    del aiocontextvars

# Expose the public API:
from ._message import Message
from ._action import (
    start_action,
    startTask,
    Action,
    preserve_context,
    current_action,
    log_call,
    log_message,
)
from ._output import ILogger, Logger, MemoryLogger, to_file, FileDestination
from ._validation import Field, fields, MessageType, ActionType, ValidationError
from ._traceback import write_traceback, writeFailure
from ._errors import register_exception_extractor
from ._version import get_versions

# Backwards compatibility:
def add_destination(destination):
    warn(
        "add_destination is deprecated since 1.1.0. " "Use add_destinations instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    Logger._destinations.add(destination)


# Backwards compatibility:
def use_asyncio_context():
    warn(
        "This function is no longer as needed as of Eliot 1.8.0.",
        DeprecationWarning,
        stacklevel=2,
    )


# Backwards compatibilty:
addDestination = add_destination
removeDestination = Logger._destinations.remove
addGlobalFields = Logger._destinations.addGlobalFields
writeTraceback = write_traceback
startAction = start_action

# PEP 8 variants:
start_task = startTask
write_failure = writeFailure
add_destinations = Logger._destinations.add
remove_destination = removeDestination
add_global_fields = addGlobalFields

# Backwards compatibility for old versions of eliot-tree, which rely on
# eliot._parse:
def _parse_compat():
    # Force eliot.parse to be imported in way that works with old Python:
    from .parse import Parser

    del Parser
    import sys

    sys.modules["eliot._parse"] = sys.modules["eliot.parse"]
    return sys.modules["eliot.parse"]


_parse = _parse_compat()
del _parse_compat


__all__ = [
    "Message",
    "writeTraceback",
    "writeFailure",
    "startAction",
    "startTask",
    "Action",
    "preserve_context",
    "Field",
    "fields",
    "MessageType",
    "ActionType",
    "ILogger",
    "Logger",
    "MemoryLogger",
    "addDestination",
    "removeDestination",
    "addGlobalFields",
    "FileDestination",
    "register_exception_extractor",
    "current_action",
    "use_asyncio_context",
    "ValidationError",
    # PEP 8 variants:
    "write_traceback",
    "write_failure",
    "start_action",
    "start_task",
    "add_destination",
    "add_destinations",
    "remove_destination",
    "add_global_fields",
    "to_file",
    "log_call",
    "log_message",
    "__version__",
    # Backwards compat for eliot-tree:
    "_parse",
]


__version__ = get_versions()["version"]
del get_versions



"""
Tests for the public API exposed by L{eliot}.
"""

from __future__ import unicode_literals

from unittest import TestCase

from .._output import Logger
import eliot


class PublicAPITests(TestCase):
    """
    Tests for the public API.
    """

    def test_addDestination(self):
        """
        L{eliot.addDestination} adds destinations to the L{Destinations}
        attached to L{Logger}.
        """
        o = object()
        eliot.addDestination(o)
        self.addCleanup(eliot.removeDestination, o)
        self.assertIn(o, Logger._destinations._destinations)

    def test_removeDestination(self):
        """
        L{eliot.addDestination} removes destinations from the L{Destinations}
        attached to L{Logger}.
        """
        self.assertEqual(eliot.removeDestination, Logger._destinations.remove)

    def test_addGlobalFields(self):
        """
        L{eliot.addGlobalFields} calls the corresponding method on the
        L{Destinations} attached to L{Logger}.
        """
        self.assertEqual(eliot.addGlobalFields, Logger._destinations.addGlobalFields)


class PEP8Tests(TestCase):
    """
    Tests for the PEP 8 variant of the the public API.
    """

    def test_add_destination(self):
        """
        L{eliot.addDestionation} is the same as L{eliot.add_destination}.
        """
        self.assertIs(eliot.add_destination, eliot.addDestination)

    def test_remove_destination(self):
        """
        L{eliot.removeDestionation} is the same as L{eliot.remove_destination}.
        """
        self.assertIs(eliot.remove_destination, eliot.removeDestination)

    def test_add_global_fields(self):
        """
        L{eliot.add_global_fields} is the same as L{eliot.addGlobalFields}.
        """
        self.assertIs(eliot.add_global_fields, eliot.addGlobalFields)

    def test_write_traceback(self):
        """
        L{eliot.writeTraceback} is the same as L{eliot.write_traceback}.
        """
        self.assertIs(eliot.write_traceback, eliot.writeTraceback)

    def test_write_failure(self):
        """
        L{eliot.writeFailure} is the same as L{eliot.write_failure}.
        """
        self.assertIs(eliot.write_failure, eliot.writeFailure)

    def test_start_task(self):
        """
        L{eliot.startTask} is the same as L{eliot.start_task}.
        """
        self.assertIs(eliot.start_task, eliot.startTask)

    def test_start_action(self):
        """
        L{eliot.startAction} is the same as L{eliot.start_action}.
        """
        self.assertIs(eliot.start_action, eliot.startAction)



"""
Tests for L{eliot.json}.
"""

from __future__ import unicode_literals, absolute_import

from unittest import TestCase, skipUnless, skipIf
from json import loads, dumps
from math import isnan

try:
    import numpy as np
except ImportError:
    np = None

from eliot.json import EliotJSONEncoder


class EliotJSONEncoderTests(TestCase):
    """Tests for L{EliotJSONEncoder}."""

    def test_nan_inf(self):
        """NaN, inf and -inf are round-tripped."""
        l = [float("nan"), float("inf"), float("-inf")]
        roundtripped = loads(dumps(l, cls=EliotJSONEncoder))
        self.assertEqual(l[1:], roundtripped[1:])
        self.assertTrue(isnan(roundtripped[0]))

    @skipUnless(np, "NumPy not installed.")
    def test_numpy(self):
        """NumPy objects get serialized to readable JSON."""
        l = [
            np.float32(12.5),
            np.float64(2.0),
            np.float16(0.5),
            np.bool(True),
            np.bool(False),
            np.bool_(True),
            np.unicode_("hello"),
            np.byte(12),
            np.short(12),
            np.intc(-13),
            np.int_(0),
            np.longlong(100),
            np.intp(7),
            np.ubyte(12),
            np.ushort(12),
            np.uintc(13),
            np.ulonglong(100),
            np.uintp(7),
            np.int8(1),
            np.int16(3),
            np.int32(4),
            np.int64(5),
            np.uint8(1),
            np.uint16(3),
            np.uint32(4),
            np.uint64(5),
        ]
        l2 = [l, np.array([1, 2, 3])]
        roundtripped = loads(dumps(l2, cls=EliotJSONEncoder))
        self.assertEqual([l, [1, 2, 3]], roundtripped)

    @skipIf(np, "NumPy is installed.")
    def test_numpy_not_imported(self):
        """If NumPy is not available, EliotJSONEncoder continues to work.

        This ensures NumPy isn't a hard dependency.
        """
        with self.assertRaises(TypeError):
            dumps([object()], cls=EliotJSONEncoder)
        self.assertEqual(dumps(12, cls=EliotJSONEncoder), "12")

    @skipUnless(np, "NumPy is not installed.")
    def test_large_numpy_array(self):
        """
        Large NumPy arrays are not serialized completely, since this is (A) a
        performance hit (B) probably a mistake on the user's part.
        """
        a1000 = np.array([0] * 10000)
        self.assertEqual(loads(dumps(a1000, cls=EliotJSONEncoder)), a1000.tolist())
        a1002 = np.zeros((2, 5001))
        a1002[0][0] = 12
        a1002[0][1] = 13
        a1002[1][1] = 500
        self.assertEqual(
            loads(dumps(a1002, cls=EliotJSONEncoder)),
            {"array_start": a1002.flat[:10000].tolist(), "original_shape": [2, 5001]},
        )



"""
Implementation of hooks and APIs for outputting log messages.
"""

import sys
import traceback
import inspect
import json as pyjson
from threading import Lock
from functools import wraps
from io import IOBase

from pyrsistent import PClass, field

from . import _bytesjson as bytesjson
from zope.interface import Interface, implementer

from ._traceback import write_traceback, TRACEBACK_MESSAGE
from ._message import EXCEPTION_FIELD, MESSAGE_TYPE_FIELD, REASON_FIELD
from ._util import saferepr, safeunicode
from .json import EliotJSONEncoder
from ._validation import ValidationError


class _DestinationsSendError(Exception):
    """
    An error occured sending to one or more destinations.

    @ivar errors: A list of tuples output from C{sys.exc_info()}.
    """

    def __init__(self, errors):
        self.errors = errors
        Exception.__init__(self, errors)


class BufferingDestination(object):
    """
    Buffer messages in memory.
    """

    def __init__(self):
        self.messages = []

    def __call__(self, message):
        self.messages.append(message)
        while len(self.messages) > 1000:
            self.messages.pop(0)


class Destinations(object):
    """
    Manage a list of destinations for message dictionaries.

    The global instance of this class is where L{Logger} instances will
    send written messages.
    """

    def __init__(self):
        self._destinations = [BufferingDestination()]
        self._any_added = False
        self._globalFields = {}

    def addGlobalFields(self, **fields):
        """
        Add fields that will be included in all messages sent through this
        destination.

        @param fields: Keyword arguments mapping field names to values.
        """
        self._globalFields.update(fields)

    def send(self, message):
        """
        Deliver a message to all destinations.

        The passed in message might be mutated.

        @param message: A message dictionary that can be serialized to JSON.
        @type message: L{dict}
        """
        message.update(self._globalFields)
        errors = []
        for dest in self._destinations:
            try:
                dest(message)
            except:
                errors.append(sys.exc_info())
        if errors:
            raise _DestinationsSendError(errors)

    def add(self, *destinations):
        """
        Adds new destinations.

        A destination should never ever throw an exception. Seriously.
        A destination should not mutate the dictionary it is given.

        @param destinations: A list of callables that takes message
            dictionaries.
        """
        buffered_messages = None
        if not self._any_added:
            # These are first set of messages added, so we need to clear
            # BufferingDestination:
            self._any_added = True
            buffered_messages = self._destinations[0].messages
            self._destinations = []
        self._destinations.extend(destinations)
        if buffered_messages:
            # Re-deliver buffered messages:
            for message in buffered_messages:
                self.send(message)

    def remove(self, destination):
        """
        Remove an existing destination.

        @param destination: A destination previously added with C{self.add}.

        @raises ValueError: If the destination is unknown.
        """
        self._destinations.remove(destination)


class ILogger(Interface):
    """
    Write out message dictionaries to some destination.
    """

    def write(dictionary, serializer=None):
        """
        Write a dictionary to the appropriate destination.

        @note: This method is thread-safe.

        @param serializer: Either C{None}, or a
            L{eliot._validation._MessageSerializer} which can be used to
            validate this message.

        @param dictionary: The message to write out. The given dictionary
             will not be mutated.
        @type dictionary: C{dict}
        """


@implementer(ILogger)
class Logger(object):
    """
    Write out messages to the globally configured destination(s).

    You will typically want to create one of these for every chunk of code
    whose messages you want to unit test in isolation, e.g. a class. The tests
    can then replace a specific L{Logger} with a L{MemoryLogger}.
    """

    _destinations = Destinations()
    _log_tracebacks = True

    def _safeUnicodeDictionary(self, dictionary):
        """
        Serialize a dictionary to a unicode string no matter what it contains.

        The resulting dictionary will loosely follow Python syntax but it is
        not expected to actually be a lossless encoding in all cases.

        @param dictionary: A L{dict} to serialize.

        @return: A L{unicode} string representing the input dictionary as
            faithfully as can be done without putting in too much effort.
        """
        try:
            return str(
                dict(
                    (saferepr(key), saferepr(value))
                    for (key, value) in dictionary.items()
                )
            )
        except:
            return saferepr(dictionary)

    def write(self, dictionary, serializer=None):
        """
        Serialize the dictionary, and write it to C{self._destinations}.
        """
        dictionary = dictionary.copy()
        try:
            if serializer is not None:
                serializer.serialize(dictionary)
        except:
            write_traceback(self)
            from ._action import log_message

            log_message(
                "eliot:serialization_failure",
                message=self._safeUnicodeDictionary(dictionary),
                __eliot_logger__=self,
            )
            return

        try:
            self._destinations.send(dictionary)
        except _DestinationsSendError as e:
            from ._action import log_message

            if self._log_tracebacks:
                for (exc_type, exception, exc_traceback) in e.errors:
                    # Can't use same Logger as serialization errors because
                    # if destination continues to error out we will get
                    # infinite recursion. So instead we have to manually
                    # construct a Logger that won't retry.
                    logger = Logger()
                    logger._log_tracebacks = False
                    logger._destinations = self._destinations
                    msg = {
                        MESSAGE_TYPE_FIELD: "eliot:destination_failure",
                        REASON_FIELD: safeunicode(exception),
                        EXCEPTION_FIELD: exc_type.__module__ + "." + exc_type.__name__,
                        "message": self._safeUnicodeDictionary(dictionary),
                        "__eliot_logger__": logger,
                    }
                    log_message(**msg)
            else:
                # Nothing we can do here, raising exception to caller will
                # break business logic, better to have that continue to
                # work even if logging isn't.
                pass


def exclusively(f):
    """
    Decorate a function to make it thread-safe by serializing invocations
    using a per-instance lock.
    """

    @wraps(f)
    def exclusively_f(self, *a, **kw):
        with self._lock:
            return f(self, *a, **kw)

    return exclusively_f


@implementer(ILogger)
class MemoryLogger(object):
    """
    Store written messages in memory.

    When unit testing you don't want to create this directly but rather use
    the L{eliot.testing.validateLogging} decorator on a test method, which
    will provide additional testing integration.

    @ivar messages: A C{list} of the dictionaries passed to
        L{MemoryLogger.write}. Do not mutate this list.

    @ivar serializers: A C{list} of the serializers passed to
        L{MemoryLogger.write}, each corresponding to a message
        L{MemoryLogger.messages}. Do not mutate this list.

    @ivar tracebackMessages: A C{list} of messages written to this logger for
        tracebacks using L{eliot.write_traceback} or L{eliot.writeFailure}. Do
        not mutate this list.
    """

    def __init__(self, encoder=EliotJSONEncoder):
        """
        @param encoder: A JSONEncoder subclass to use when encoding JSON.
        """
        self._lock = Lock()
        self._encoder = encoder
        self.reset()

    @exclusively
    def flushTracebacks(self, exceptionType):
        """
        Flush all logged tracebacks whose exception is of the given type.

        This means they are expected tracebacks and should not cause the test
        to fail.

        @param exceptionType: A subclass of L{Exception}.

        @return: C{list} of flushed messages.
        """
        result = []
        remaining = []
        for message in self.tracebackMessages:
            if isinstance(message[REASON_FIELD], exceptionType):
                result.append(message)
            else:
                remaining.append(message)
        self.tracebackMessages = remaining
        return result

    # PEP 8 variant:
    flush_tracebacks = flushTracebacks

    @exclusively
    def write(self, dictionary, serializer=None):
        """
        Add the dictionary to list of messages.
        """
        # Validate copy of the dictionary, to ensure what we store isn't
        # mutated.
        try:
            self._validate_message(dictionary.copy(), serializer)
        except Exception as e:
            # Skip irrelevant frames that don't help pinpoint the problem:
            from . import _output, _message, _action

            skip_filenames = [_output.__file__, _message.__file__, _action.__file__]
            for frame in inspect.stack():
                if frame[1] not in skip_filenames:
                    break
            self._failed_validations.append(
                "{}: {}".format(e, "".join(traceback.format_stack(frame[0])))
            )
        self.messages.append(dictionary)
        self.serializers.append(serializer)
        if serializer is TRACEBACK_MESSAGE._serializer:
            self.tracebackMessages.append(dictionary)

    def _validate_message(self, dictionary, serializer):
        """Validate an individual message.

        As a side-effect, the message is replaced with its serialized contents.

        @param dictionary: A message C{dict} to be validated.  Might be mutated
            by the serializer!

        @param serializer: C{None} or a serializer.

        @raises TypeError: If a field name is not unicode, or the dictionary
            fails to serialize to JSON.

        @raises eliot.ValidationError: If serializer was given and validation
            failed.
        """
        if serializer is not None:
            serializer.validate(dictionary)
        for key in dictionary:
            if not isinstance(key, str):
                if isinstance(key, bytes):
                    key.decode("utf-8")
                else:
                    raise TypeError(dictionary, "%r is not unicode" % (key,))
        if serializer is not None:
            serializer.serialize(dictionary)

        try:
            pyjson.dumps(dictionary, cls=self._encoder)
        except Exception as e:
            raise TypeError("Message %s doesn't encode to JSON: %s" % (dictionary, e))

    @exclusively
    def validate(self):
        """
        Validate all written messages.

        Does minimal validation of types, and for messages with corresponding
        serializers use those to do additional validation.

        As a side-effect, the messages are replaced with their serialized
        contents.

        @raises TypeError: If a field name is not unicode, or the dictionary
            fails to serialize to JSON.

        @raises eliot.ValidationError: If serializer was given and validation
            failed.
        """
        for dictionary, serializer in zip(self.messages, self.serializers):
            try:
                self._validate_message(dictionary, serializer)
            except (TypeError, ValidationError) as e:
                # We already figured out which messages failed validation
                # earlier. This just lets us figure out which exception type to
                # raise.
                raise e.__class__("\n\n".join(self._failed_validations))

    @exclusively
    def serialize(self):
        """
        Serialize all written messages.

        This is the Field-based serialization, not JSON.

        @return: A C{list} of C{dict}, the serialized messages.
        """
        result = []
        for dictionary, serializer in zip(self.messages, self.serializers):
            dictionary = dictionary.copy()
            serializer.serialize(dictionary)
            result.append(dictionary)
        return result

    @exclusively
    def reset(self):
        """
        Clear all logged messages.

        Any logged tracebacks will also be cleared, and will therefore not
        cause a test failure.

        This is useful to ensure a logger is in a known state before testing
        logging of a specific code path.
        """
        self.messages = []
        self.serializers = []
        self.tracebackMessages = []
        self._failed_validations = []


class FileDestination(PClass):
    """
    Callable that writes JSON messages to a file.

    On Python 3 the file may support either C{bytes} or C{unicode}.  On
    Python 2 only C{bytes} are supported since that is what all files expect
    in practice.

    @ivar file: The file to which messages will be written.

    @ivar _dumps: Function that serializes an object to JSON.

    @ivar _linebreak: C{"\n"} as either bytes or unicode.
    """

    file = field(mandatory=True)
    encoder = field(mandatory=True)
    _dumps = field(mandatory=True)
    _linebreak = field(mandatory=True)

    def __new__(cls, file, encoder=EliotJSONEncoder):
        if isinstance(file, IOBase) and not file.writable():
            raise RuntimeError("Given file {} is not writeable.")

        unicodeFile = False
        try:
            file.write(b"")
        except TypeError:
            unicodeFile = True

        if unicodeFile:
            # On Python 3 native json module outputs unicode:
            _dumps = pyjson.dumps
            _linebreak = "\n"
        else:
            _dumps = bytesjson.dumps
            _linebreak = b"\n"
        return PClass.__new__(
            cls, file=file, _dumps=_dumps, _linebreak=_linebreak, encoder=encoder
        )

    def __call__(self, message):
        """
        @param message: A message dictionary.
        """
        self.file.write(self._dumps(message, cls=self.encoder) + self._linebreak)
        self.file.flush()


def to_file(output_file, encoder=EliotJSONEncoder):
    """
    Add a destination that writes a JSON message per line to the given file.

    @param output_file: A file-like object.

    @param encoder: A JSONEncoder subclass to use when encoding JSON.
    """
    Logger._destinations.add(FileDestination(file=output_file, encoder=encoder))


# The default Logger, used when none is specified:
_DEFAULT_LOGGER = Logger()



"""Test for pyinstaller compatibility."""

from __future__ import absolute_import

from unittest import TestCase, SkipTest
from tempfile import mkdtemp, NamedTemporaryFile
from subprocess import check_call, CalledProcessError
import os

from six import PY2

if PY2:
    FileNotFoundError = OSError


class PyInstallerTests(TestCase):
    """Make sure PyInstaller doesn't break Eliot."""

    def setUp(self):
        try:
            check_call(["pyinstaller", "--help"])
        except (CalledProcessError, FileNotFoundError):
            raise SkipTest("Can't find pyinstaller.")

    def test_importable(self):
        """The Eliot package can be imported inside a PyInstaller packaged binary."""
        output_dir = mkdtemp()
        with NamedTemporaryFile(mode="w") as f:
            f.write("import eliot; import eliot.prettyprint\n")
            f.flush()
            check_call(
                [
                    "pyinstaller",
                    "--distpath",
                    output_dir,
                    "-F",
                    "-n",
                    "importeliot",
                    f.name,
                ]
            )
        check_call([os.path.join(output_dir, "importeliot")])



"""
Tests for L{eliot._parse}.
"""

from __future__ import unicode_literals

from unittest import TestCase
from itertools import chain

from six import text_type as unicode, assertCountEqual
from six.moves import zip_longest

from hypothesis import strategies as st, given, assume

from pyrsistent import PClass, field, pvector_field

from .. import start_action, Message
from ..testing import MemoryLogger
from ..parse import Task, Parser
from .._message import (
    WrittenMessage,
    MESSAGE_TYPE_FIELD,
    TASK_LEVEL_FIELD,
    TASK_UUID_FIELD,
)
from .._action import FAILED_STATUS, ACTION_STATUS_FIELD, WrittenAction
from .strategies import labels


class ActionStructure(PClass):
    """
    A tree structure used to generate/compare to Eliot trees.

    Individual messages are encoded as a unicode string; actions are
    encoded as a L{ActionStructure} instance.
    """

    type = field(type=(unicode, None.__class__))
    children = pvector_field(object)  # XXX ("StubAction", unicode))
    failed = field(type=bool)

    @classmethod
    def from_written(cls, written):
        """
        Create an L{ActionStructure} or L{unicode} from a L{WrittenAction} or
        L{WrittenMessage}.
        """
        if isinstance(written, WrittenMessage):
            return written.as_dict()[MESSAGE_TYPE_FIELD]
        else:  # WrittenAction
            if not written.end_message:
                raise AssertionError("Missing end message.")
            return cls(
                type=written.action_type,
                failed=(
                    written.end_message.contents[ACTION_STATUS_FIELD] == FAILED_STATUS
                ),
                children=[cls.from_written(o) for o in written.children],
            )

    @classmethod
    def to_eliot(cls, structure_or_message, logger):
        """
        Given a L{ActionStructure} or L{unicode}, generate appropriate
        structured Eliot log mesages to given L{MemoryLogger}.
        """
        if isinstance(structure_or_message, cls):
            action = structure_or_message
            try:
                with start_action(logger, action_type=action.type):
                    for child in action.children:
                        cls.to_eliot(child, logger)
                    if structure_or_message.failed:
                        raise RuntimeError("Make the eliot action fail.")
            except RuntimeError:
                pass
        else:
            Message.new(message_type=structure_or_message).write(logger)
        return logger.messages


@st.composite
def action_structures(draw):
    """
    A Hypothesis strategy that creates a tree of L{ActionStructure} and
    L{unicode}.
    """
    tree = draw(st.recursive(labels, st.lists, max_leaves=20))

    def to_structure(tree_or_message):
        if isinstance(tree_or_message, list):
            return ActionStructure(
                type=draw(labels),
                failed=draw(st.booleans()),
                children=[to_structure(o) for o in tree_or_message],
            )
        else:
            return tree_or_message

    return to_structure(tree)


def _structure_and_messages(structure):
    messages = ActionStructure.to_eliot(structure, MemoryLogger())
    return st.permutations(messages).map(lambda permuted: (structure, permuted))


# Hypothesis strategy that creates a tuple of ActionStructure/unicode and
# corresponding serialized Eliot messages, randomly shuffled.
STRUCTURES_WITH_MESSAGES = action_structures().flatmap(_structure_and_messages)


def parse_to_task(messages):
    """
    Feed a set of messages to a L{Task}.

    @param messages: Sequence of messages dictionaries to parse.

    @return: Resulting L{Task}.
    """
    task = Task()
    for message in messages:
        task = task.add(message)
    return task


class TaskTests(TestCase):
    """
    Tests for L{Task}.
    """

    @given(structure_and_messages=STRUCTURES_WITH_MESSAGES)
    def test_missing_action(self, structure_and_messages):
        """
        If we parse messages (in shuffled order) but a start message is
        missing then the structure is still deduced correctly from the
        remaining messages.
        """
        action_structure, messages = structure_and_messages
        assume(not isinstance(action_structure, unicode))

        # Remove first start message we encounter; since messages are
        # shuffled the location removed will differ over Hypothesis test
        # iterations:
        messages = messages[:]
        for i, message in enumerate(messages):
            if message[TASK_LEVEL_FIELD][-1] == 1:  # start message
                del messages[i]
                break

        task = parse_to_task(messages)
        parsed_structure = ActionStructure.from_written(task.root())

        # We expect the action with missing start message to otherwise
        # be parsed correctly:
        self.assertEqual(parsed_structure, action_structure)

    @given(structure_and_messages=STRUCTURES_WITH_MESSAGES)
    def test_parse_from_random_order(self, structure_and_messages):
        """
        If we shuffle messages and parse them the parser builds a tree of
        actions that is the same as the one used to generate the messages.

        Shuffled messages means we have to deal with (temporarily) missing
        information sufficiently well to be able to parse correctly once
        the missing information arrives.
        """
        action_structure, messages = structure_and_messages

        task = Task()
        for message in messages:
            task = task.add(message)

        # Assert parsed structure matches input structure:
        parsed_structure = ActionStructure.from_written(task.root())
        self.assertEqual(parsed_structure, action_structure)

    @given(structure_and_messages=STRUCTURES_WITH_MESSAGES)
    def test_is_complete(self, structure_and_messages):
        """
        ``Task.is_complete()`` only returns true when all messages within the
        tree have been delivered.
        """
        action_structure, messages = structure_and_messages

        task = Task()
        completed = []
        for message in messages:
            task = task.add(message)
            completed.append(task.is_complete())

        self.assertEqual(completed, [False for m in messages[:-1]] + [True])

    def test_parse_contents(self):
        """
        L{{Task.add}} parses the contents of the messages it receives.
        """
        logger = MemoryLogger()
        with start_action(logger, action_type="xxx", y=123) as ctx:
            Message.new(message_type="zzz", z=4).write(logger)
            ctx.add_success_fields(foo=[1, 2])
        messages = logger.messages
        expected = WrittenAction.from_messages(
            WrittenMessage.from_dict(messages[0]),
            [WrittenMessage.from_dict(messages[1])],
            WrittenMessage.from_dict(messages[2]),
        )

        task = parse_to_task(messages)
        self.assertEqual(task.root(), expected)


class ParserTests(TestCase):
    """
    Tests for L{Parser}.
    """

    @given(
        structure_and_messages1=STRUCTURES_WITH_MESSAGES,
        structure_and_messages2=STRUCTURES_WITH_MESSAGES,
        structure_and_messages3=STRUCTURES_WITH_MESSAGES,
    )
    def test_parse_into_tasks(
        self, structure_and_messages1, structure_and_messages2, structure_and_messages3
    ):
        """
        Adding messages to a L{Parser} parses them into a L{Task} instances.
        """
        _, messages1 = structure_and_messages1
        _, messages2 = structure_and_messages2
        _, messages3 = structure_and_messages3
        all_messages = (messages1, messages2, messages3)
        # Need unique UUIDs per task:
        assume(len(set(m[0][TASK_UUID_FIELD] for m in all_messages)) == 3)

        parser = Parser()
        all_tasks = []
        for message in chain(*zip_longest(*all_messages)):
            if message is not None:
                completed_tasks, parser = parser.add(message)
                all_tasks.extend(completed_tasks)

        assertCountEqual(
            self, all_tasks, [parse_to_task(msgs) for msgs in all_messages]
        )

    @given(structure_and_messages=STRUCTURES_WITH_MESSAGES)
    def test_incomplete_tasks(self, structure_and_messages):
        """
        Until a L{Task} is fully parsed, it is returned in
        L{Parser.incomplete_tasks}.
        """
        _, messages = structure_and_messages
        parser = Parser()
        task = Task()
        incomplete_matches = []
        for message in messages[:-1]:
            _, parser = parser.add(message)
            task = task.add(message)
            incomplete_matches.append(parser.incomplete_tasks() == [task])

        task = task.add(messages[-1])
        _, parser = parser.add(messages[-1])
        self.assertEqual(
            dict(
                incomplete_matches=incomplete_matches,
                final_incompleted=parser.incomplete_tasks(),
            ),
            dict(incomplete_matches=[True] * (len(messages) - 1), final_incompleted=[]),
        )

    @given(
        structure_and_messages1=STRUCTURES_WITH_MESSAGES,
        structure_and_messages2=STRUCTURES_WITH_MESSAGES,
        structure_and_messages3=STRUCTURES_WITH_MESSAGES,
    )
    def test_parse_stream(
        self, structure_and_messages1, structure_and_messages2, structure_and_messages3
    ):
        """
        L{Parser.parse_stream} returns an iterable of completed and then
        incompleted tasks.
        """
        _, messages1 = structure_and_messages1
        _, messages2 = structure_and_messages2
        _, messages3 = structure_and_messages3
        # Need at least one non-dropped message in partial tree:
        assume(len(messages3) > 1)
        # Need unique UUIDs per task:
        assume(
            len(set(m[0][TASK_UUID_FIELD] for m in (messages1, messages2, messages3)))
            == 3
        )

        # Two complete tasks, one incomplete task:
        all_messages = (messages1, messages2, messages3[:-1])

        all_tasks = list(
            Parser.parse_stream(
                [m for m in chain(*zip_longest(*all_messages)) if m is not None]
            )
        )
        assertCountEqual(
            self, all_tasks, [parse_to_task(msgs) for msgs in all_messages]
        )


class BackwardsCompatibility(TestCase):
    """Tests for backwards compatibility."""

    def test_imports(self):
        """Old ways of importing still work."""
        import eliot._parse
        from eliot import _parse
        import eliot.parse

        self.assertIs(eliot.parse, eliot._parse)
        self.assertIs(_parse, eliot.parse)
